/*
 * This file is part of din.
 *
 * din is copyright (c) 2006 - 2011 S Jagannathan <jag@dinisnoise.org>
 * For more information, please visit http://dinisnoise.org
 *
 * din is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 2 of the License, or
 * (at your option) any later version.
 *
 * din is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 * for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with din.  If not, see <http://www.gnu.org/licenses/>.
 *
*/

#include "audio.h"
#include "chrono.h"
#include "command.h"
#include "console.h"
#include "tcl_interp.h"
#include <math.h>
#include <fstream>
#include <limits.h>
#include <stdlib.h>
using namespace std;

extern int SAMPLE_RATE;
extern double SECS_PER_BUFFER;

string name ("din"), dinl, dinr;
static const char* colon = ":", *L = "L", * R = "R", *MIDI_IN = "midi";

extern string dotdin;

audio_out::audio_out () {

	prefs_name = dotdin + "jack_prefs";
	load_prefs ();
	open ();
}

audio_out::~audio_out () {

	close ();

	float* ptr [] = {samples_buffers, step, left, right, result, ams, fms};
	for (int i = 0; i < 7; ++i) delete[] ptr[i];

	delete[] available;

	jack_ringbuffer_free(midi_rb);
}

void audio_out::alloc () {

	samples_per_buffer = num_channels * samples_per_channel;
	samples_buffer_size = sizeof (float) * samples_per_buffer;

	if (samples_buffers) delete[] samples_buffers;
	samples_buffers = new float [num_samples_buffers * samples_per_buffer];

	if (available) delete[] available;
	available = new int [num_samples_buffers];

	readi = writei = 0;
	readp = writep = samples_buffers;
	for (int i = 0; i < num_samples_buffers; ++i) available[i] = 0;

	float** ptr [] = {&step, &left, &right, &result, &ams, &fms};
	for (int i = 0; i < 6; ++i) {
		if (ptr[i]) delete[] *ptr[i];
		*ptr[i] = new float [samples_per_channel];
	}

	midi_rb = jack_ringbuffer_create(samples_buffer_size);
}

void audio_out::set_sample_rate (int s) {

	SAMPLE_RATE = sample_rate = s;
	SECS_PER_BUFFER = samples_per_channel * 1. / SAMPLE_RATE;
	calc_tonic_limits ();
}

int audio_out::open() {

	cout << "<<< connecting to JACK server >>>" << endl;
	jack_status_t status;
	jack_options_t options = JackNullOption;

	client = jack_client_open (name.c_str(), options, &status, 0);
	if (client == 0) {
		cout << "!!! couldnt connect to JACK server. !!!" << endl;
		exit (1);
	} else {
		name = string (jack_get_client_name (client));
		cout << "+++ connected to JACK server as: " << name << " +++" << endl;
		dinl = name + colon + L;
		dinr = name + colon + R;
	}

	sample_rate = jack_get_sample_rate (client);
	jack_set_sample_rate_callback (client, sample_rate_changed, this);

	samples_per_channel = jack_get_buffer_size (client);
	jack_set_buffer_size_callback (client, buffer_size_changed, this);

	out_left = jack_port_register (client, L, JACK_DEFAULT_AUDIO_TYPE, JackPortIsOutput, 0);
	out_right = jack_port_register (client, R, JACK_DEFAULT_AUDIO_TYPE, JackPortIsOutput, 0);
	if ((out_left == 0) || (out_right == 0)) {
		cout << "!!! couldnt create output ports !!!" << endl;
		exit (0);
	} else cout << "+++ opened L and R for output +++" << endl;
	lcons = rcons = 0;

	midi_in = jack_port_register (client, MIDI_IN, JACK_DEFAULT_MIDI_TYPE, JackPortIsInput, 0);
	if (midi_in == NULL) {
		cout << "!!! couldn't create midi input port !!!" << endl;
		exit (0);
	}

	alloc ();
	set_sample_rate (sample_rate);

	return 0;

}

int audio_out::close () {
	if (client) jack_client_close (client);
	return 0;
}

int audio_out::start () {

	if (client) {

		jack_set_process_callback (client, audio_wanted, this);
		jack_activate (client);

		if (lcons) { // restore connections of L
			for (int i = 0; i < lcons; ++i) jack_connect (client, dinl.c_str(), con_left[i]);
			free (con_left);
		} else if (jack_connect (client, dinl.c_str(), "system:playback_1") != 0) {
			cout << "!!! couldnt connect to playback_1 !!!" << endl;
		}

		if (rcons) { // restore connections of R
			for (int i = 0; i < lcons; ++i) jack_connect (client, dinr.c_str(), con_right[i]);
			free (con_right);
		} else if (jack_connect (client, dinr.c_str(), "system:playback_2") != 0) {
			cout << "!!! couldnt connect to playback_2 !!!" << endl;
		}

		cout << "+++ streaming audio now +++ " << endl;

	}
	return 0;
}

int audio_out::stop () {
	if (client) {
		lcons = jack_port_connected (out_left);
		rcons = jack_port_connected (out_right);
		if (lcons) con_left = jack_port_get_connections (out_left);
		if (rcons) con_right = jack_port_get_connections (out_right);
		jack_deactivate (client);
	}
	return 0;
}

void audio_out::write (float* sbuf, float* lbuf, float* rbuf) {

	for (int i = 0; i < samples_per_buffer; i += 2) {
		*lbuf++ = *sbuf++;
		*rbuf++ = *sbuf++;
	}

}

void audio_out::load_prefs () {

	ifstream file (prefs_name.c_str(), ios::in);
	string ignore;

	if (!file)
		defaults ();
	else {
		string type;
		file >> ignore >> type;
		if (type != "audio_prefs") {
			cout << "!!! bad audio prefrences file, please check: " << prefs_name << ", will use defaults. !!!" << eol;
			num_samples_buffers = 2;
		} else {
			cout << "<< loading audio prefs from: " << prefs_name;
			file >> ignore >> num_samples_buffers;
			cout << ", done. >>>" << eol;
		}
	}
}

void audio_out::save_prefs () {
	ofstream file (prefs_name.c_str(), ios::out);
	if (file) {
		file << "type: audio_prefs" << eol;
		file << "num_samples_buffers: " << num_samples_buffers << eol;
	} else cout << "!!! couldnt write file: " << prefs_name << " !!!" << eol;
}

void audio_out::defaults () {
	cout << "<<< using default for JACK settings >>>" << endl;
	num_samples_buffers = 2;
}

extern audio_out aout;
extern chrono clk;

int buffer_size_changed (jack_nframes_t nframes, void *arg) {

	if ((unsigned int) aout.samples_per_channel != jack_get_buffer_size (aout.client)) {
		cout << "!!! buffer size change not supported, please restart din !!!" << endl;
	}
	return 0;

}

int sample_rate_changed (jack_nframes_t nframes, void *arg) {
	aout.set_sample_rate (jack_get_sample_rate(aout.client));
	return 0;
}

int audio_wanted (jack_nframes_t nsamples, void *arg) {

	if (aout.available[aout.readi]) {
		sample_t* lbuf = (sample_t *) jack_port_get_buffer (aout.out_left, aout.samples_per_channel); // left buf
		sample_t* rbuf = (sample_t *) jack_port_get_buffer (aout.out_right, aout.samples_per_channel); // right buf
		aout.write (aout.readp, lbuf, rbuf);
		aout.available [aout.readi] = 0;
		if (++aout.readi >= aout.num_samples_buffers) {
			aout.readi = 0;
			aout.readp = aout.samples_buffers;
		} else aout.readp += aout.samples_per_buffer;
		clk += aout.samples_per_buffer;
	}

	void *midi_buf = jack_port_get_buffer (aout.midi_in, nsamples);
	jack_nframes_t event_count = jack_midi_get_event_count (midi_buf);

	if (event_count > 0) {

		jack_nframes_t frame_start = jack_last_frame_time (aout.client);
		jack_nframes_t event_index;
		for (event_index = 0; event_index < event_count; ++event_index) {
			jack_midi_event_t ev;
			jack_midi_event_get (&ev, midi_buf, event_index);

			// Push event onto ringbuffer
			if (jack_ringbuffer_write_space (aout.midi_rb) < sizeof ev.size + sizeof ev.time + ev.size) break;

			ev.time += frame_start;
			jack_ringbuffer_write (aout.midi_rb, (const char *)&ev.size, sizeof ev.size);
			jack_ringbuffer_write (aout.midi_rb, (const char *)&ev.time, sizeof ev.time);
			jack_ringbuffer_write (aout.midi_rb, (const char *)ev.buffer, ev.size);
		}
	}

	return 0;
}

static void run (const string& c, size_t length, jack_midi_data_t *buffer)
{
	char buf[10];
	unsigned char *bufp = (unsigned char *)buffer;
	string cmd (c);

	for (size_t i = 0; i < length; i++) {
		sprintf(buf, " %u", *bufp++);
		cmd += buf;
	}
	cons (cmd);
}

void process_jackmidi ()
{

	extern double MIDI_BPM;
	extern tcl_interp interpreter;

	static double start_time = 0;
	static int num_clocks = 0;

	jack_midi_event_t ev;
	jack_midi_data_t* ebuf = 0;
	unsigned int ebuf_size = 0;

	while (1) {

		size_t f = jack_ringbuffer_read_space (aout.midi_rb);
		if (f < sizeof ev.size) break;

		jack_ringbuffer_peek (aout.midi_rb, (char *) &ev.size, sizeof ev.size);
		if (f < sizeof ev.size + sizeof ev.time + ev.size) break;

		int alloc = 0;
		if (ebuf) {
			if (ev.size > ebuf_size) {
				free (ebuf);
				alloc = 1;
			}
		} else alloc = 1;
		if (alloc) {
			ebuf = (jack_midi_data_t *) malloc (ev.size);
			ebuf_size = ev.size;
		}

		ev.buffer = ebuf;
		jack_ringbuffer_read(aout.midi_rb, (char *)&ev.size, sizeof ev.size);
		jack_ringbuffer_read(aout.midi_rb, (char *)&ev.time, sizeof ev.time);
		jack_ringbuffer_read(aout.midi_rb, (char *)ev.buffer, ev.size);

		if ((*ev.buffer & 0xf0) == 0xb0) { // control change
			run ("midi-cc", ev.size, ev.buffer);
		} else if ((*ev.buffer & 0xf0) == 0x90) { // note on
			run ("midi-note-on", ev.size, ev.buffer);
		} else if ((*ev.buffer & 0xf0) == 0x80) { // note off
			run ("midi-note-off", ev.size, ev.buffer);
		} else if ((*ev.buffer & 0xf0) == 0xc0) { // program change
			run ("midi-program-change", ev.size, ev.buffer);
		} else if ((*ev.buffer & 0xf0) == 0xe0) { // pitch bend

			char status = ev.buffer[0];
			char lsb = ev.buffer[1];
			char msb = ev.buffer[2];

			// Normalise pitchbend range to -1.0...1.0
			short value = (lsb | (msb << 7)) - 0x2000;
			float fvalue = (float)value / ((value < 0) ? 0x2000 : 0x1FFF);

			ostringstream msg;
			msg << "midi-pitch-bend " << status << " " << value << " " << fvalue;
			cons (msg.str());

		} else if (*ev.buffer == 0xf8) { // midi clock
			double elapsed_time;
			if (ev.time >= start_time) {
				elapsed_time = ev.time - start_time;
			} else { // Time has wrapped (~ 27 hours @ 44100Hz)
				elapsed_time = UINT_MAX - start_time + ev.time;
			}
			++num_clocks;

			if (elapsed_time >= SAMPLE_RATE) {
				double f_num_clocks = 1. / elapsed_time * num_clocks;
				double midi_bpm = f_num_clocks * 2.5 * SAMPLE_RATE;
				MIDI_BPM = midi_bpm;
				Tcl_UpdateLinkedVar (interpreter.interp, "midibpm");
				cons ("midi-clock");
				num_clocks = 0;
				start_time = ev.time;
			}
		} else if (*ev.buffer == 0xfa) { // midi start
			cons ("midi-start");
			num_clocks = 0;
			start_time = ev.time;
		}
	}

	if (ebuf) free(ebuf);
}
